/******************************************************************************
Connection.java

CVS Information:

  $Source: /homes/dsl/Master/cif/CIF-Comm-Java-Current/libraries/cif/comm/Connection.java,v $
  $Date: 1998/10/13 06:33:46 $
  $Revision: 1.3 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import cif.toolkit.*;

/**
 * An abstract interface for a CIF Connection object.  New connection objects
 * are created using the CIF Factory class.
 *
 * @see cif.comm.Factory
 */
public
interface Connection
{
    /**
     * Enable the connection to deliver events to the event listener.  Note,
     * this is necessary to avoid race conditions associated with events
     * arriving before the application is capable of dealing with them.  If
     * events occur before this method is called, the events are queued and
     * delivered once event notification is enabled.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     */
    public
    void
    enable_Event_Notification()
    throws
	Object_Inactive_Exception;
	
    /**
     * Stop delivering events to the event listener.  Note, this method will
     * block until all in-progress event dispatching associated with this
     * connection has completed.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     */
    public
    void
    disable_Event_Notification()
    throws
	Object_Inactive_Exception;
	
    /**
     * Send the specified message data to the connection object(s)
     * participating in this connection.  This method simply calls
     * cif.comm.Connection.send(byte[], int, int) with the message_tag set to
     * zero.
     *
     * @see cif.comm.Connection#send(byte[], int, int)
     */
    public
    void
    send(
	byte []				message_data,
	int				message_length)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception;

    
    /**
     * Send the specified message data to the connection object(s)
     * participating in this connection.  The subset of participating
     * connection objects which receive the message is dependent on the
     * communication protocol associated with the connection.
     *
     * @param message_data
     *
     *     A byte array containing the message to be sent.
     *
     * @param message_length
     *
     *     The number of bytes to send.  This number must not exceed the length
     *     of the byte array.
     *
     * @param message_tag
     *
     *     An integer identifier for this message.  Message tags exist only as
     *     a convenience to the user.  They are not interpreted in any way by
     *     the CIF communication package.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the messages sent by this object may not have been
     *     delivered to the other objects participating in the connection, and
     *     that some of the messages sent by those objects may not have been
     *     received by this object.  However, the exact implications of the
     *     failure are dependent on the underlying protocol.
     *
     * @exception cif.comm.Comm_Flow_Exception
     *
     *      Comm_Flow_Exception is thrown when the connection does not support
     *      an outgoing flow.
     */
    public
    void
    send(
	byte []				message_data,
	int				message_length,
	int				message_tag)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception;
    

    
    /**
     * Obtain the next message received on this connection.
     *
     * @see cif.comm.Connection#receive(cif.toolkit.Timeout)
     */
    public
    Message
    receive()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    
    /**
     * Obtain the next message received on this connection.
     *
     * @see cif.comm.Connection#receive(cif.toolkit.Timeout)
     */
    public
    Message
    receive(
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;

    /**
     * Obtain the next message received on this connection.  If no messages
     * are immediately available, this method will block until one of the
     * following conditions are met <UL>
     *
     *     <LI> a new message arrives,
     *
     *     <LI> the specified amount of time elapses,
     *
     *     <LI> the connection is closed,
     *
     *     <LI> the thread is interrupted, or
     *
     *     <LI> some problem is detected with the underlying communication
     *     channel. </UL>
     *
     * @param timeout
     *
     *     If the next message is not immediately available, wait up to point
     *     where the timeout object expires for it to arrive.
     *
     * @return
     *
     *     The received message is return in a CIF Message object.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.toolkit.Timeout_Expired_Exception
     *
     *     Timeout_Expired_Exception is thrown if the specified amount of time
     *     elapses before a message arrives.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the messages sent by this object may not have been
     *     delivered to the other objects participating in the connection, and
     *     that some of the messages sent by those objects may not have been
     *     received by this object.  However, the exact implications of the
     *     failure are dependent on the underlying protocol.
     *
     * @exception cif.comm.Comm_Flow_Exception
     *
     *      Comm_Flow_Exception is thrown when the connection does not support
     *      an incoming flow.
     *
     * @exception cif.comm.Comm_Would_Block_Exception
     *
     *     Comm_Would_Block_Exception is thrown if the receive() method was
     *     called by the same thread currently handling a CIF event, and the
     *     receive() method would have blocked waiting for a message to arrive.
     *     If called from an event handler, receive() may only be called with a
     *     timeout value of zero.
     *
     * @exception InterruptedException
     *
     *     InterruptedException is thrown if the thread was interrupted for
     *     some reason.  Presumably, the user's code intentionally performed
     *     the interrupt, so the exception is passed through.
     *
     */
    public
    Message
    receive(
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;


    /**
     * Check the connection for newly arrived messages.
     *
     * @return
     *
     *     If a message is waiting to be received, then a value of true is
     *     returned; otherwise, false is returned.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the messages sent by this object may not have been
     *     delivered to the other objects participating in the connection, and
     *     that some of the messages sent by those objects may not have been
     *     received by this object.  However, the exact implications of the
     *     failure are dependent on the underlying protocol.
     *
     * @exception cif.comm.Comm_Flow_Exception
     *
     *      Comm_Flow_Exception is thrown when the connection does not support
     *      an incoming flow.
     */
    public
    boolean
    poll()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception;

    /**
     * Close the connection.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the messages sent by this object may not have been
     *     delivered to the other objects participating in the connection, and
     *     that some of the messages sent by those objects may not have been
     *     received by this object.  However, the exact implications of the
     *     failure are dependent on the underlying protocol.
     */
    public
    void
    close()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception;
}
