/******************************************************************************
I_Connection.java

CVS Information:

  $Source: /homes/dsl/Master/cif/CIF-Comm-Java-Current/libraries/cif/comm/I_Connection.java,v $
  $Date: 1998/10/13 20:25:45 $
  $Revision: 1.4 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import cif.toolkit.*;
import nexus.*;


/**
 * This class is an implementation of the CIF Comm Connection class for Nexus
 * protocols.  This class does not implement the connection establishment,
 * however.  That functionality is left to protocol specific subclasses.
 *
 * @see cif.comm.Connection
 */
abstract class I_Connection
implements
    Connection,
    Event_Listener
{
    /**
     * Initialize the internal object state.  To activate the object, call one
     * of the defined activate() methods.
     */
    I_Connection()
    {
	l_ref_cnt = new I_Ref_Count();
	l_reset_Internal_State();
    }
    /* I_Connection.I_Connection() */

    
    /**
     * Establish a new connection using the specified URL.
     */
    synchronized
    void
    activate(
	URL					url,
	Connection_Properties			properties,
	Timeout					timeout)
    throws
	Object_Active_Exception,
	Timeout_Expired_Exception,
	Bad_Parameter_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	InterruptedException
    {
	/*
	 * Verify that this is not already a valid connection object
	 */
	l_verify_Object_Inactive();
	    
	/*
	 * Verify URL information
	 */
	if (url.get_Hostname() == null)
	{
	    throw new Bad_Parameter_Exception(
		1, "hostname must be specified in a connection URL");
	}

	if (url.get_Port() < 1)
	{
	    throw new Bad_Parameter_Exception(
		1, "port number may not be zero in a connection URL");
	}

	/*
	 * Establish a new connection
	 */
	I_Endpoint_Properties			ep_properties;
	boolean					perform_reset;

	perform_reset = true;
	
	try
	{
	    ep_properties = new I_Endpoint_Properties();
	    ep_properties.set_Event_Listener(this);
	    open_Connection(url, ep_properties, timeout);

	    perform_reset = false;
	}
	finally
	{
	    if (perform_reset)
	    {
		l_reset_Internal_State();
	    }
	}
	

	/*
	 * Set the connection state to indicate that it is active
	 */
	l_event_listener = properties.get_Event_Listener();
	l_active = true;
    }
    /* activate() */


    /**
     * Establish a new connection using the startpoint in the specified buffer
     */
    synchronized
    void
    activate(
	GetBuffer				nx_buf,
	Connection_Properties			properties,
	Timeout					timeout)
    throws
	Object_Active_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	InterruptedException
    {
	/*
	 * Verify that this is not already a valid connection object
	 */
	l_verify_Object_Inactive();
	    
	/*
	 * Establish a new connection
	 */
	I_Endpoint_Properties			ep_properties;
	boolean					perform_reset;

	perform_reset = true;
	
	try
	{
	    ep_properties = new I_Endpoint_Properties();
	    ep_properties.set_Event_Listener(this);
	    open_Connection(nx_buf, ep_properties, timeout);

	    perform_reset = false;
	}
	finally
	{
	    if (perform_reset)
	    {
		l_reset_Internal_State();
	    }
	}

	/*
	 * Set the connection state to indicate that it is active
	 */
	l_event_listener = properties.get_Event_Listener();
	l_active = true;
    }
    /* activate() */


    /**
     * Enable event notification
     *
     * @see cif.comm.Connection#enable_Event_Notification()
     */
    public
    void
    enable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	I_Endpoint				ep;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();

	    /*
	     * Verify that this connection supports an outgoing flow
	     */
	    if ((i_flow & I_FLOW_INCOMING) == 0)
	    {
		return;
	    }
	    
	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = i_ep;
	}

	try
	{
	    ep.enable_Event_Notification();
	}
	catch(Exception e)
	{
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    /* enable_Event_Notification() */

    
    /**
     * Disable event notification
     *
     * @see cif.comm.Connection#disable_Event_Notification()
     */
    public
    void
    disable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	I_Endpoint				ep;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();

	    /*
	     * Verify that this connection supports an outgoing flow
	     */
	    if ((i_flow & I_FLOW_INCOMING) == 0)
	    {
		return;
	    }
	    
	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = i_ep;
	}

	try
	{
	    ep.disable_Event_Notification();
	}
	catch(Exception e)
	{
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    /* disable_Event_Notification() */

    
    /**
     * Send the specified message data with a message tag of zero to the
     * endpoint(s) participating in this connection.
     *
     * @see cif.comm.Connection#send(byte[], int)
     */
    public
    void
    send(
	byte []					message_data,
	int					message_length)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception
    {
	l_send(message_data, message_length, 0);
    }
    /* I_Connection.send(byte[], int) */


    /**
     * Send the specified message data with the specified message tag to the
     * endpoint(s) participating in this connection.
     *
     * @see cif.comm.Connection#send(byte[], int)
     */
    public
    void
    send(
	byte []					message_data,
	int					message_length,
	int					message_tag)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception
    {
	l_send(message_data, message_length, message_tag);
    }
    /* I_Connection.send(byte[], int, int) */

    
    /**
     * Internal routine to send a message to remote endpoints associated with
     * this connection.
     */
    private
    void
    l_send(
	byte []					message_data,
	int					message_length,
	int					message_tag)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception
    {
	Nexus					nx;
	I_Startpoint				sp;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();
	    l_verify_No_Errors();

	    /*
	     * Verify that this connection supports an outgoing flow
	     */
	    if ((i_flow & I_FLOW_OUTGOING) == 0)
	    {
		throw new Comm_Flow_Exception();
	    }

	    /*
	     * Obtain a refence to Nexus so that we can create a buffer
	     */
	    nx = Comm.acquire_Nexus();
	    
	    /*
	     * Obtain a reference to the startpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    sp = i_sp;
	}
	
	try
	{
	    /*
	     * Pack the message into a Nexus buffer
	     */
	    PutBuffer				buffer;
	    int					buffer_sz;

	    buffer_sz =
		DataConversion.sizeof_int(2) +
		DataConversion.sizeof_byte(message_length);

	    buffer = nx.buffer_init(buffer_sz, 0);
	    Comm.release_Nexus();
	
	    try
	    {
		buffer.put_int(message_length);
		buffer.put_int(message_tag);
		buffer.put_byte(message_data, 0, message_length);
	    }
	    catch (BufferOverrunException e)
	    {
		throw new Internal_Error(
		    "I_Connection.send() caught an unexpected " +
		    "BufferOverrunException");
	    }

	    /*
	     * Send the buffer to all remote endpoints associated with this
	     * connection
	     */
	    sp.send(buffer);
	}
	finally
	{
	    /*
	     * Decrement reference count, notifying waiting threads that we are
	     * have finished referencing members of this startpoint object
	     */
	    l_ref_cnt.dec_Count();
	}
    }
    /* I_Connection.l_send() */

    
    /**
     * Obtain the next message in the receive queue.  If the queue is currently
     * empty, then wait for a message to arrive.
     *
     * @see cif.comm.Connection#receive()
     */
    public
    Message
    receive()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	Timeout					timeout_tmp;
	
	timeout_tmp = new Timeout(Timeout.INFINITY);
	try
	{
	    return l_receive(timeout_tmp);
	}
	catch (Timeout_Expired_Exception e)
	{
	    throw new Internal_Error(
		"I_Connection.receive() caught an unexpected " +
		"Timeout_Expired_Exception");
	}
    }
    /* I_Connection.receive() */
    

    /**
     * Obtain the next message in the receive queue.  If the queue is currently
     * empty, then wait up to the specified amount of time for a message to
     * arrive.
     *
     * @see cif.comm.Connection#receive(int)
     */
    public
    Message
    receive(
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	Timeout					timeout_tmp;
	
	timeout_tmp = new Timeout(timeout);
	return l_receive(timeout_tmp);
    }
    /* I_Connection.receive(long) */
    
    /**
     * Obtain the next message in the receive queue.  If the queue is currently
     * empty, then wait up to the specified amount of time for a message to
     * arrive.
     *
     * @see cif.comm.Connection#receive(cif.toolkit.Timeout)
     */
    public
    Message
    receive(
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	return l_receive(timeout);
    }
    /* I_Connection.receive(cif.toolkit.Timeout) */
    
    /**
     * Internal routine which return the next message in the message queue.
     */
    private
    Message
    l_receive(
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	I_Endpoint				ep;
	GetBuffer				buffer;
	Message					message;
	Message.Info				message_info;

	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();
	    l_verify_No_Errors();

	    /*
	     * Verify that this connection supports an incoming flow
	     */
	    if ((i_flow & I_FLOW_INCOMING) == 0)
	    {
		throw new Comm_Flow_Exception();
	    }
	    
	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = i_ep;
	}
	
	try
	{
	    /*
	     * Obtain the next message from the queue
	     */
	    buffer = ep.receive(timeout);
	}
	finally
	{
	    /*
	     * Decrement reference count, notifying waiting threads that we are
	     * have finished referencing members of this startpoint object
	     */
	    l_ref_cnt.dec_Count();
	}
	
	/*
	 * Fill in the message meta-information object
	 */
	message_info = new Message.Info();
	
	try
	{
	    message_info.size = buffer.get_int();
	    message_info.tag = buffer.get_int();
	}
	catch (Exception e)
	{
	    throw new Internal_Error(
		"I_Connection.receive(): encountered an unexpected " +
		"buffer extraction problem");
	}

	/*
	 * Create a new message object and extract the data into it
	 */
	message = new Message(message_info);

	try
	{
	    buffer.get_byte(message.get_Data(), 0, message_info.size);
	}
	catch (Exception e)
	{
	    throw new Internal_Error(
		"I_Connection.receive() encountered an unexpected " +
		"buffer extraction problem");
	}

	/*
	 * Return the message to the user
	 */
	return message;
    }
    /* l_receive() */


    /**
     * Check the connection for newly arrived messages.
     *
     * @see cif.comm.Connection#poll()
     */
    public
    boolean
    poll()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Flow_Exception
    {
	I_Endpoint				ep;
	boolean					rc;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();
	    l_verify_No_Errors();
	    
	    /*
	     * Verify that this connection supports an incoming flow
	     */
	    if ((i_flow & I_FLOW_INCOMING) == 0)
	    {
		throw new Comm_Flow_Exception();
	    }
	    
	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = i_ep;
	}
	
	try
	{
	    /*
	     * Check for newly arrived messages
	     */
	    rc = ep.poll();
	}
	finally
	{
	    /*
	     * Decrement reference count, notifying waiting threads that we are
	     * have finished referencing members of this startpoint object
	     */
	    l_ref_cnt.dec_Count();
	}
	
	return rc;
    }
    /* poll() */
    

    /**
     * Close the connection by deactivating the object.
     *
     * @see cif.comm.Connection#close()
     */
    public
    void
    close()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	deactivate();
    }
    /* close() */

    
    /**
     * Disable the connection.  This method simply calls the deactivate()
     * method in a last ditch effort to properly deactivate the connection
     * object before it is deleted.
     */
    public
    void
    finalize()
    {
	try
	{
	    deactivate();
	}
	catch (Exception e)
	{
	}
    }
    /* I_Connection.finalize() */
    

    /**
     * Tear down an existing connection
     */
    void 
    deactivate()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();
	
	    if (!l_deactivate_pending)
	    {
		/*
		 *Deactivate the NSML startpoint and endpoint objects
		 */
		try
		{
		    if ((i_flow & I_FLOW_OUTGOING) != 0)
		    {
			i_sp.deactivate();
		    }

		    if ((i_flow & I_FLOW_INCOMING) != 0)
		    {
			i_ep.deactivate();
		    }
		}
		catch (Object_Inactive_Exception e)
		{
		    throw new Internal_Error(
			"I_Connection.deactivate() enountered problems " +
			"deactivating its internal startpoint/endpoint pair");
		}

		l_deactivate_pending = true;
	    }
	}

	l_ref_cnt.wait_Until_Zero();

	synchronized(this)
	{
	    boolean				errors;
	    
	    /*
	     * Since another thread may have already finished the deactivation,
	     * we need to verify that the object is still active
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    };
	    
	    errors = l_error;

	    l_reset_Internal_State();

	    if (errors)
	    {
		throw new Comm_Failure_Exception();
	    }
	}
    }
    /* deactivate() */

    
    /**************************************************************************
    **************************************************************************/

    protected static final int			I_FLOW_UNESTABLISHED = 0;
    protected static final int			I_FLOW_OUTGOING = 1;
    protected static final int			I_FLOW_INCOMING = 2;
    protected static final int			I_FLOW_BIDIRECTIONAL = 3;
    
    /**
     * Establish a new connection using the specified URL.
     */
    protected abstract
    void
    open_Connection(
	URL					url,
	I_Endpoint_Properties			ep_properties,
	Timeout					timeout)
    throws
	Bad_Parameter_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	InterruptedException;
    
    
    /**
     * Establish a new connection using the startpoint of a listener.
     */
    protected abstract
    void
    open_Connection(
	GetBuffer				nx_buf,
	I_Endpoint_Properties			ep_properties,
	Timeout					timeout)
    throws
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	InterruptedException;
    
    
    /**************************************************************************
    **************************************************************************/

    /**
     * Handle incoming events from the endpoint associated with this connection
     */
    public
    void
    handle_Event(
	Event					ep_event)
    {
	Event_Listener				event_listener;
	Connection_Message_Arrival_Event	conn_event;
	
	synchronized(this)
	{
	    if (!l_active)
	    {
		throw new Internal_Error(
		    "cif.comm.I_Connection.handle_Event: receive an event " +
		    "while in the inactive state");
	    }

	    event_listener = l_event_listener;
	    l_ref_cnt.inc_Count();
	}

	try
	{
	    conn_event = new Connection_Message_Arrival_Event(this);
	    event_listener.handle_Event(conn_event);
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    /* I_Connection.handle_Event() */

    
    /**
     * Verify that the object has been activated
     */
    private
    void
    l_verify_Object_Active()
    throws
	Object_Inactive_Exception
    {
	if (!l_active)
	{
	    throw new Object_Inactive_Exception();
	}
    }
    /* l_verify_Object_Active() */
    

    /**
     * Verify that the object has not already been activated
     */
    private
    void
    l_verify_Object_Inactive()
    throws
	Object_Active_Exception
    {
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}
    }
    /* l_verify_Object_Inactive() */
    

    /**
     * Verify that the object has observed any communication failures
     */
    private
    void
    l_verify_No_Errors()
    throws
	Comm_Failure_Exception
    {
	if (l_error)
	{
	    throw new Comm_Failure_Exception();
	}
    }
    /* l_verify_No_Errors() */
    

    /**
     * Reset the internal state of the connection object
     */
    private
    void
    l_reset_Internal_State()
    {
	l_error = false;
	i_flow = I_FLOW_UNESTABLISHED;
	i_ep = null;
	i_sp = null;
	l_deactivate_pending = false;
	l_active = false;
    }
    /* I_Connection.l_reset_Internal_State() */

    
    /**************************************************************************
    **************************************************************************/

    
    // Object state flag (true --> active, false --> inactive)
    private boolean				l_active;

    // Object error flag (true --> error, false --> no error)
    private boolean				l_error;

    // Data flow(s) support by this connection
    int						i_flow;
    
    // NSML startpoint and endpoint object used to form a connection
    I_Startpoint				i_sp;
    I_Endpoint					i_ep;

    // Event handling information
    private Event_Listener			l_event_listener;
    
    /* synchronization variables to ensure deactivation doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    private I_Ref_Count				l_ref_cnt;
    private boolean				l_deactivate_pending;
}
