/******************************************************************************
I_Endpoint.java

CVS Information:

  $Source: /homes/dsl/Master/cif/CIF-Comm-Java-Current/libraries/cif/comm/I_Endpoint.java,v $
  $Date: 1998/10/13 20:25:46 $
  $Revision: 1.3 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import nexus.*;
import cif.toolkit.*;

/**
 * This class defines an endpoint of a Nexus Simple Messaging Layer (NSML)
 * connection.  The endpoint is used to synchronously receive messages sent
 * from an NSML startpoint.
 */
class I_Endpoint
implements
    nexus.HandlerInterface,
    nexus.AttachApprovalInterface
{
    /**************************************************************************
     *************************************************************************/
    
    /**
     * Initialize internal endpoint object state.  To activate an endpoint
     * object, call one of the activate() methods.
     */
    public
    I_Endpoint()
    {
	l_ref_cnt = new I_Ref_Count();
	l_reset_Internal_State();
    }
    /* I_Endpoint.I_Endpoint() */

    
    /**
     * Enable the endpoint to being receiving messages using the specified
     * protocol.
     *
     * @param properties
     *
     *     Properties to be asssociated with this endpoint.
     *
     * @exception cif.toolkit.Object_Active_Exception
     *
     *     Object_Active_Exception is thrown if the object has been previously
     *     activated.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying communication
     *     structures failed to initialize.  This could be caused by specifying
     *     an incorrect protoocol, or because enabling remote attach processing
     *     failed.
     */
    public synchronized
    void
    activate(
	I_Endpoint_Properties			properties)
    throws
	Object_Active_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	nexus.Nexus				nx;
	
	/*
	 * Verify that the object is not currently active
	 */
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}

	/*
	 * Initialize event listener information
	 */
	l_event.listener = properties.get_Event_Listener();
	
	/*
	 * Get a reference to Nexus
	 */
	nx = Comm.acquire_Nexus();
	
	/*
	 * Setup an appropriate set of attributes for the endpoint and
	 * then intiialize the endpoint
	 */
	nexus.EndpointAttr			nx_ep_attr;
	
	nx_ep_attr = nx.endpointattr_init();
	nx_ep_attr.set_handler_object(this);
	nx_ep_attr.set_protocol(properties.get_Protocol());

	l_nx_ep = nx.endpoint_init(nx_ep_attr);
	l_nx_ep.set_user_object(this);

	/* If the application has request attach capabilities, then setup Nexus
	 * to process attach requests
	 */
	if (properties.is_Attach_Enabled())
	{
	    l_attach.port =
		nx.allow_attach(properties.get_Attach_Port(), this, this);
	    if (l_attach.port < 0)
	    {
		l_attach.port = 0;
		l_nx_ep.destroy();
		l_nx_ep = null;

		Comm.release_Nexus();
		
		throw new Comm_Failure_Exception(
		    "I_Endpoint.activate(): nexus.allow_attach() failed");
	    }
	    
	    l_attach.hostname = nx.attach_hostname();
	    l_attach.enabled = true;
	}

	Comm.release_Nexus();
	
	/*
	 * Set the object state
	 */
	l_active = true;
    }
    /* I_Endpoint.activate() */

    
    /**
     * Enable the endpoint to deliver events to the event listener.  Note, this
     * is necessary to avoid race conditions associated with events arriving
     * before the application is capable of dealing with them.  If events occur
     * before this method is called, the events are queued and delivered once
     * event notification is enabled.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     */
    public
    void
    enable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    synchronized(l_event)
	    {
		if (l_event.listener == null)
		{
		    return;
		}
		
		if (l_event.state != L_Event_Info.STATE_DISABLED)
		{
		    return;
		}

		l_event.state = L_Event_Info.STATE_ENABLING;
	    }
		    
	    l_ref_cnt.inc_Count();
	}
	

	try
	{
	    while (true)
	    {
		L_Queue_Entry			entry;
		cif.toolkit.Event_Listener		event_listener;
	    
		synchronized(this)
		{
		    /*
		     * Verify that the endpoint is still
		     */
		    if (l_deactivate_pending)
		    {
			throw new Object_Inactive_Exception();
		    }

		    synchronized(l_event)
		    {
			l_event.dispatching = false;
		    
			if (l_event.state == L_Event_Info.STATE_DISABLED)
			{
			    l_event.notifyAll();
			    break;
			}
		    
			entry = l_q_head;
			while(entry != null)
			{
			    if (entry.event_dispatched == false)
			    {
				break;
			    }

			    entry = entry.next;
			}

			if (entry == null)
			{
			    if (l_event.state == L_Event_Info.STATE_ENABLING)
			    {
				l_event.state = L_Event_Info.STATE_ENABLED;
			    }
			
			    break;
			}

			event_listener = l_event.listener;
			l_event.dispatching = true;
			l_event.thread = Thread.currentThread();
		    }
		    // end synchronized(l_event)

		    entry.event_dispatched = true;
		}
		// end synchronized(this)
	    
		I_Endpoint_Message_Arrival_Event	event;

		event = new I_Endpoint_Message_Arrival_Event(this);
		event_listener.handle_Event(event);
	    }
	    // end while(true)
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    /* I_Endpoint.enable_Event_Notification() */
    
    /**
     * Stop delivering events to the event listener.  Note, this method will
     * block until all in-progress event dispatching associated with this
     * endpoint has completed.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     */
    public
    void
    disable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	boolean					must_wait;
	
	must_wait = false;
		
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    synchronized(l_event)
	    {
		if (l_event.listener == null)
		{
		    return;
		}

		if (l_event.state != L_Event_Info.STATE_DISABLED)
		{
		    l_event.state = L_Event_Info.STATE_DISABLED;
		    
		    if (l_event.thread != Thread.currentThread())
		    {
			l_ref_cnt.inc_Count();
			must_wait = true;
		    }
		}
	    }
	}

	if (must_wait) 
	{
	    try
	    {
		synchronized(l_event)
		{
		    while (l_event.dispatching)
		    {
			try
			{
			    l_event.wait();
			}
			catch (InterruptedException e)
			{
			    break;
			}

			if (l_deactivate_pending)
			{
			    break;
			}
		    }
		}
	    }
	    finally
	    {
		l_ref_cnt.dec_Count();
	    }
	}
    }
    /* I_Endpoint.disable_Event_Notification() */
    

    /**
     * Obtain the next message received by this endpoint object.  If the next
     * message is not immediately available, wait until one of the
     * following conditions is met: <UL>
     *
     *     <LI> a message arrives,
     *
     *     <LI> the specified amount of time elapses,
     *
     *     <LI> the endpoint is deactivated, or
     *
     *     <LI> the thread is interrupted </UL>
     *
     * @param timeout
     *
     *     If the next message is not immediately available, wait up to timeout
     *     number of millseconds for it to arrive.
     *
     * @return
     *
     *     The received message is returned in a Nexus GetBuffer object.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     * @exception cif.toolkit.Timeout_Expired_Exception
     *
     *     Timeout_Expired_Exception is thrown if the specified amount of time
     *     elapses before a message arrives.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Would_Block_Exception
     *
     *     Comm_Would_Block_Exception is thrown if receive() is being called
     *	   from the event handler associated with this object.
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying communication
     *     link fails for some reason (network outage, etc.)
     *
     * @exception InterruptedException
     *
     *     InterruptedException is thrown if the thread was interrupted for
     *     some reason.  Presumably, the user's code intentionally performed
     *     the interrupt, so the exception is passed through.
     */
    public synchronized
    nexus.GetBuffer
    receive(
	Timeout					timeout)
    throws
        Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	/*
	 * Verify that the object is a valid, active endpoint
	 */
	if (!l_active || l_deactivate_pending)
	{
	    throw new Object_Inactive_Exception();
	}
	
	/*
	 * If the message queue is empty, then wait for a new message to arrive
	 */
	if (l_q_head == null)
	{
	    synchronized(l_event)
	    {
		if (l_event.dispatching == true &&
		    l_event.thread == Thread.currentThread())
		{
		    throw new Comm_Would_Block_Exception();
		}
	    }
	    
	    while(true)
	    {
		/*
		 * Verify that the endpoint isn't being deactivated
		 */
		if (l_deactivate_pending)
		{
		    throw new Object_Inactive_Exception();
		}
		
		/*
		 * Note: referencing count is placed around the blocking wait()
		 * so that the object will not be prematurely deactivated.
		 */
		l_ref_cnt.inc_Count();

		try
		{
		    if (timeout.is_Infinity())
		    {
			wait();
		    }
		    else
		    {
			wait(timeout.get(), 1);
		    }
		}
		finally
		{
		    l_ref_cnt.dec_Count();
		}

		/*
		 * If an new message has arrived then stop waiting
		 */
		if(l_q_head != null)
		{
		    break;
		}

		/*
		 * If the user supplied timeout has elapsed, then return
		 * control to the application
		 */
		if (timeout.is_Expired())
		{
		    throw new Timeout_Expired_Exception();
		}
	    }
	    // end while(true)
	}
	// end if (l_q_head == null)

	/*
	 * Dequeue the message and return a reference to it
	 */
	L_Queue_Entry				entry;

	entry = l_q_head;
	l_q_head = entry.next;
	if (l_q_head == null)
	{
	    l_q_tail = null;
	}
	
	return entry.buf;
    }
    /* I_Endpoint.receive() */

    
    /**
     * Check the endpoint for newly arrived messages.
     *
     * @return
     *
     *     If a message is waiting to be received, then a value of true is
     *     returned; otherwise, false is returned.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     */
    public synchronized
    boolean
    poll()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Verify that the object has been activated
	 */
	if (!l_active)
	{
	    throw new Object_Inactive_Exception();
	}
	
	return (l_q_head != null);
    }
    /* I_Endpoint.poll() */


    /**
     * Get the name for the current host which is listening for remote attach
     * requests.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     */
    public synchronized
    String
    get_Attach_Hostname()
    throws
	Object_Inactive_Exception
    {
	/*
	 * Verify that the object has been activated
	 */
	if (!l_active || !l_attach.enabled)
	{
	    throw new Object_Inactive_Exception();
	}

	return l_attach.hostname;
    }
    
	
    /**
     * Get the port number that the current host is expect to receive remote
     * attach requests on.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     */
    public synchronized
    int
    get_Attach_Port()
    throws
	Object_Inactive_Exception
    {
	/*
	 * Verify that the object has been activated
	 */
	if (!l_active || !l_attach.enabled)
	{
	    throw new Object_Inactive_Exception();
	}

	return l_attach.port;
    }
    
	
    /**
     * Disable the endpoint.  The underlying communcation channel is closed
     * and any messages remaining in the queue are discarded.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     */
    public
    void
    deactivate()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Initiate deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    /*
	     * Get access to the Nexus object
	     */
	    nexus.Nexus			nx;
		
	    nx = Comm.acquire_Nexus();
	    
	    if (!l_deactivate_pending)
	    {
		/*
		 * Notify all blocking threads that this object is going away
		 */
		l_deactivate_pending = true;
	    	notifyAll();

		/*
		 * If Nexus is listening for remote attach requests, then
		 * shutdown the listener
		 */
		nx.disallow_attach(l_attach.port);
	    }

	    Comm.release_Nexus();
	}
	// end synchronized(this)
	    
	/*
	 * Wait for blocking operations to complete
	 */
	l_ref_cnt.wait_Until_Zero();

	/*
	 * Finsh deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Since another thread may have already finished the deactivation,
	     * we need to verify that the object is still active
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    /*
	     * Destroy the endpoint
	     */
	    l_nx_ep.destroy();

	    /*
	     * Drain the message queue
	     */
	    l_q_head = null;
	    l_q_tail = null;
	    
	    /*
	     * Reset the object's internal state
	     */
	    l_reset_Internal_State();
	}
	// end synchronized(this)
    }
    /* I_Endpoint.deactivate() */


    /**
     * Disable the endpoint.  This method simply calls the deactivate() method
     * in a last ditch effort to properly deactivate the endpoint object before
     * it is deleted.
     */
    public
    void
    finalize()
    {
	try
	{
	    deactivate();
	}
	catch (Exception e)
	{
	}
    }
    /* I_Endpoint.finalize() */

    
    /**************************************************************************
    **************************************************************************/

    /**
     * Obtain the internal Nexus endpoint object.  This used by other objects
     * which are part of the CIF package.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     */
    synchronized
    Endpoint
    i_get_Nexus_Endpoint()
    throws
	Object_Inactive_Exception
    {
	/*
	 * Verify that the object has been activated
	 */
	if (!l_active)
	{
	    throw new Object_Inactive_Exception();
	}
	
	return l_nx_ep;
    }
    /* I_Endpoint.i_get_Nexus_Endpoint() */

    
    /**
     * Handler method to receive incoming messages from Nexus.
     */
    public synchronized
    void
    invoke_handler(
	Endpoint				ep,
	GetBuffer				buf,
	int					hid)
    {
	cif.toolkit.Event_Listener		event_listener;
	
	synchronized(this)
	{
	    if (!l_active)
	    {
		throw new Internal_Error(
		    "cif.comm.I_Endpoint.invoke_handler(): " +
		    "message recieved after object was deactivated");
	    }

	    /*
	     * Add the buffer to the message queue and wakeup one of the
	     * waiting receivers
	     */
	    L_Queue_Entry			entry;

	    entry = new L_Queue_Entry();
	    
	    entry.buf = buf;
	    entry.event_dispatched = false;
	    entry.next = null;
	    
	    if (l_q_tail == null)
	    {
		l_q_head = entry;
	    }
	    else
	    {
		l_q_tail.next = entry;
	    }
	    l_q_tail = entry;
	    notify();

	    /*
	     * If the application has an event handler registered with the
	     * endpoint, then obtain a reference to the event listener and bump
	     * the number of outstanding operations.
	     *
	     * NOTE: since we are using non-threaded handlers, it shouldn't be
	     * possible for multiple event dispatches to occur simultaneously,
	     * hence a dispatch count and thread specific state are not
	     * required.
	     */
	    event_listener = null;
	    synchronized(l_event)
	    {
		if (l_event.state == L_Event_Info.STATE_ENABLED &&
		    l_event.listener != null)
		{
		    event_listener = l_event.listener;
		    l_event.dispatching = true;
		    l_event.thread = Thread.currentThread();

		    entry.event_dispatched = true;

		    l_ref_cnt.inc_Count();
		}
	    }
	    // end synchronized(l_event)
	}
	// end synchronized(this)

	/*
	 * If an event listener is registered then dispatch a message arrival
	 * event
	 */
	if (event_listener != null)
	{
	    try
	    {
		I_Endpoint_Message_Arrival_Event	event;

		event = new I_Endpoint_Message_Arrival_Event(this);
		event_listener.handle_Event(event);

		/*
		 * If the event listener has been disabled while we were
		 * dispatching an event, we must wake up the thread(s) that
		 * called disable_Event_Listener().
		 */
		synchronized(l_event)
		{
		    l_event.dispatching = false;
		    if (l_event.state != L_Event_Info.STATE_ENABLED)
		    {
			l_event.notifyAll();
		    }
		}
		// end synchronized(l_event)
	    }
	    finally
	    {
		l_ref_cnt.dec_Count();
	    }
	}
    }
    /* I_Endpoint.invoke_handler() */


    /**
     * Handler method to process incoming attach requests.
     */
    public
    nexus.AttachReturn
    attach_approval(
	Object					user_arg,
	String					url)
    {
        Startpoint				sp;
	nexus.Nexus				nx;
	
	if (!l_active)
	{
	    throw new Internal_Error(
		"cif.comm.I_Endpoint.attach_approval(): " +
		"attach request recieved after object was deactivated");
	}

	try
	{
	    nx = Comm.acquire_Nexus();
	}
	catch (Comm_Inactive_Exception e)
	{
	    throw new Internal_Error(
		"cif.comm.I_Endpoint.attach_approval(): " +
		"attach request recieved after Nexus was deactivated");
	}
	
        sp = nx.startpoint_bind(l_nx_ep);

	Comm.release_Nexus();
	
        return (new AttachReturn(0, sp));
    }
    /* I_Endpoint.attach_approval() */

    /**
     * I_Endpoint.l_reset_Internal_State()
     */
    private
    void
    l_reset_Internal_State()
    {
	l_q_head = null;
	l_q_tail = null;
	l_attach = new L_Attach_Info();
	l_event = new L_Event_Info();
	l_nx_ep = null;
	l_deactivate_pending = false;
	l_active = false;
    }
    /* I_Endpoint.l_reset_Internal_State() */
    
    
    /**************************************************************************
     *************************************************************************/
    
    private static
    class L_Queue_Entry
    {
	nexus.GetBuffer				buf;
	boolean					event_dispatched;
	L_Queue_Entry				next;
    }

    private static
    class L_Event_Info
    {
	static final int			STATE_DISABLED = 0;
	static final int			STATE_ENABLING = 1;
	static final int			STATE_ENABLED = 2;
	
	L_Event_Info()
	{
	    listener = null;
	    state = STATE_DISABLED;
	    dispatching = false;
	    thread = null;
	}
	
	cif.toolkit.Event_Listener		listener;
	int					state;
	boolean					dispatching;
	Thread					thread;
    }

    private static
    class L_Attach_Info
    {
	L_Attach_Info()
	{
	    enabled = false;
	    port = 0;
	    hostname = null;
	}
	
	boolean					enabled;
	int					port;
	String					hostname;
    }
    
    /**************************************************************************
    **************************************************************************/

    // Object state flag (true --> active, false --> inactive)
    private boolean				l_active;

    /* synchronization variables to ensure deactivation doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    private I_Ref_Count				l_ref_cnt;
    private volatile boolean			l_deactivate_pending;

    // Nexus endpoint for receiving incoming messages
    private nexus.Endpoint			l_nx_ep;

    // Event dispatching information
    private L_Event_Info			l_event;
    
    // Remote attach information
    private L_Attach_Info			l_attach;
    
    // Incoming message queue
    private L_Queue_Entry			l_q_head;
    private L_Queue_Entry			l_q_tail;
}
