/******************************************************************************
I_Listener.java

CVS Information:

  $Source: /homes/dsl/Master/cif/CIF-Comm-Java-Current/libraries/cif/comm/I_Listener.java,v $
  $Date: 1998/10/13 20:25:46 $
  $Revision: 1.3 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import cif.toolkit.*;
import nexus.*;

/**
 * This class is an implementation of the CIF Comm Listener class for Nexus
 * protocols.  This class does not implement the connection establishment,
 * however.  That functionality is left to protocol specific subclasses.
 *
 * @see cif.comm.Listener
 */
abstract class I_Listener
implements
    Listener,
    Event_Listener
{
    /**
     * Initialize the internal object state.  To activate the object, call one
     * of the defined activate() methods.
     */
    I_Listener()
    {
	l_ref_cnt = new I_Ref_Count();
	l_reset_Internal_State();
    }
    /* I_Listener.I_Listener() */


    synchronized
    void
    activate(
	URL					url,
	Listener_Properties			properties)
    throws
	Object_Active_Exception,
	Bad_Parameter_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Verify that this is not already a valid listener object
	 */
	l_verify_Object_Inactive();

	/*
	 * Verify URL information and get port number
	 */
	int					port;
	
	if (url.get_Hostname() != null)
	{
	    throw new Bad_Parameter_Exception(
		1, "hostname should not be specified in a listener URL");
	}

	port = url.get_Port();
	
	/*
	 * Establish a new listener
	 */
	boolean					perform_reset;

	perform_reset = true;
	
	try
	{
	    /*
	     * Create an endpoint for this connection
	     */
	    I_Endpoint_Properties		ep_properties;

	    ep_properties = new I_Endpoint_Properties();
	    ep_properties.set_Event_Listener(this);
	    ep_properties.enable_Attach();
	    ep_properties.set_Attach_Port(port);

	    l_ep = new I_Endpoint();
	    l_ep.activate(ep_properties);
	    
	    /*
	     * Form a URL for this connection
	     */
	    try
	    {
		l_url = get_URL_Scheme() + "://" +
		    l_ep.get_Attach_Hostname() + ":"
		    + l_ep.get_Attach_Port();
	    }
	    catch (Object_Inactive_Exception e)
	    {
		throw new Internal_Error(
		    "I_Connection.active() caught an unexpected " +
		    "Object_Inactive_Exception while forming the URL");
	    }
	    
		
	    perform_reset = false;
	}
	finally
	{
	    if (perform_reset)
	    {
		l_reset_Internal_State();
	    }
	}

	/*
	 * Set the connection state to indicate that it is active
	 */
	l_event_listener = properties.get_Event_Listener();
	l_active = true;
    }
    
    
    /**
     * Enable the listener to deliver events to the event listener.  Note,
     * this is necessary to avoid race conditions associated with events
     * arriving before the application is capable of dealing with them.  If
     * events occur before this method is called, the events are queued and
     * delivered once event notification is enabled.
     */
    public
    void
    enable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	I_Endpoint				ep;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();

	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = l_ep;
	}

	try
	{
	    ep.enable_Event_Notification();
	}
	catch(Exception e)
	{
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    
	
    /**
     * Stop delivering events to the event listener.  Note, this method will
     * block until all in-progress event dispatching associated with this
     * listener has completed.
     */
    public
    void
    disable_Event_Notification()
    throws
	Object_Inactive_Exception
    {
	I_Endpoint				ep;
	
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();

	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = l_ep;
	}

	try
	{
	    ep.disable_Event_Notification();
	}
	catch(Exception e)
	{
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    
	
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established.
     *
     * @see cif.comm.Listener#get_Connection()
     */
    public
    Connection
    get_Connection()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	try
	{
	    return l_get_Connection(new Connection_Properties(),
				    new Timeout(Timeout.INFINITY));
	}
	catch (Timeout_Expired_Exception e)
	{
	    throw new Internal_Error(
		"I_Connection.receive() caught an unexpected " +
		"Timeout_Expired_Exception");
	}
    }
    
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(long)
     */
    public
    Connection
    get_Connection(
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	return l_get_Connection(new Connection_Properties(),
				new Timeout(timeout));
    }
    
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	return l_get_Connection(new Connection_Properties(),
				timeout);
    }
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties)
     */
    public
    Connection
    get_Connection(
        Connection_Properties			properties)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	try
	{
	    return l_get_Connection(properties,
				    new Timeout(Timeout.INFINITY));
	}
	catch (Timeout_Expired_Exception e)
	{
	    throw new Internal_Error(
		"I_Connection.receive() caught an unexpected " +
		"Timeout_Expired_Exception");
	}
    }
    
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, long)
     */
    public
    Connection
    get_Connection(
	Connection_Properties			properties,
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	return l_get_Connection(properties,
				new Timeout(timeout));
    }
    
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties,
     *                                       cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	Connection_Properties			properties,
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	return l_get_Connection(properties, timeout);
    }
    
    /**
     * Internal routine to establish a connection using the next message on the
     * receive queue
     */
    private
    Connection
    l_get_Connection(
	Connection_Properties			properties,
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException
    {
	I_Endpoint				ep;
	GetBuffer				nx_buf;

	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();

	    /*
	     * Obtain a reference to the endpoint associated with this object
	     */
	    l_ref_cnt.inc_Count();
	    ep = l_ep;
	}
	
	try
	{
	    /*
	     * Obtain the next message from the queue
	     */
	    nx_buf = ep.receive(timeout);

	    /*
	     * Attempt to open a new connection using the startpoint contained
	     * in the message
	     */
	    return open_Connection(nx_buf, properties, timeout);
	}
	finally
	{
	    /*
	     * Decrement reference count, notifying waiting threads that we are
	     * have finished referencing members of this startpoint object
	     */
	    l_ref_cnt.dec_Count();
	}
    }
    

    /**
     * Obtain the URL which can be used to form a new connection.
     */
    public synchronized
    String
    get_URL()
    throws
	Object_Inactive_Exception
    {
	/*
	 * Verify that the object is valid and that no communication errors
	 * have occurred
	 */
	l_verify_Object_Active();

	return l_url;
    }

    
    /**
     * Close (deactivate) the listener.
     */
    public
    void
    close()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception
    {
	deactivate();
    }

    
    /**
     * Disable the listener.  This method simply calls the deactivate()
     * method in a last ditch effort to properly deactivate the listener
     * object before it is deleted.
     */
    public
    void
    finalize()
    {
	try
	{
	    deactivate();
	}
	catch (Exception e)
	{
	}
    }
    /* I_Listener.finalize() */
    
    /**
     * Tear down an existing listener
     */
    void 
    deactivate()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception
    {
	synchronized(this)
	{
	    /*
	     * Verify that the object is valid and that no communication errors
	     * have occurred
	     */
	    l_verify_Object_Active();
	
	    if (!l_deactivate_pending)
	    {
		/*
		 *Deactivate the NSML startpoint and endpoint objects
		 */
		try
		{
		    l_ep.deactivate();
		}
		catch (Object_Inactive_Exception e)
		{
		    throw new Internal_Error(
			"I_Connection.deactivate() enountered problems " +
			"deactivating its internal startpoint/endpoint pair");
		}
		catch(Comm_Failure_Exception e)
		{
		}
		
		l_deactivate_pending = true;
	    }
	}

	l_ref_cnt.wait_Until_Zero();

	synchronized(this)
	{
	    boolean				errors;
	    
	    /*
	     * Since another thread may have already finished the deactivation,
	     * we need to verify that the object is still active
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    };
	    
	    l_reset_Internal_State();
	}
    }
    /* deactivate() */

    

    /**************************************************************************
    **************************************************************************/

    /**
     * Establish a new connection using the startpoint of a listener.
     */
    protected abstract
    Connection
    open_Connection(
	GetBuffer				nx_buf,
	Connection_Properties			properties,
	Timeout					timeout)
    throws
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	InterruptedException;

    protected abstract
    String
    get_URL_Scheme();
    
    
    /**************************************************************************
    **************************************************************************/

    /**
     * Handle incoming events from the endpoint associated with this connection
     */
    public
    void
    handle_Event(
	Event					ep_event)
    {
	Event_Listener				event_listener;
	Listener_Connection_Request_Event	listener_event;
	
	synchronized(this)
	{
	    if (!l_active)
	    {
		throw new Internal_Error(
		    "cif.comm.I_Listener.handle_Event: receive an event " +
		    "while in the inactive state");
	    }

	    event_listener = l_event_listener;
	    l_ref_cnt.inc_Count();
	}

	try
	{
	    listener_event = new Listener_Connection_Request_Event(this);
	    event_listener.handle_Event(listener_event);
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }
    /* I_Listener.handle_Event() */

    
    /**
     * Verify that the object has been activated
     */
    private
    void
    l_verify_Object_Active()
    throws
	Object_Inactive_Exception
    {
	if (!l_active)
	{
	    throw new Object_Inactive_Exception();
	}
    }
    /* l_verify_Object_Active() */
    

    /**
     * Verify that the object has not already been activated
     */
    private
    void
    l_verify_Object_Inactive()
    throws
	Object_Active_Exception
    {
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}
    }
    /* l_verify_Object_Inactive() */
    

    /**
     * Reset the internal state of the listener object
     */
    private
    void
    l_reset_Internal_State()
    {
	l_url = null;
	l_event_listener = null;
	l_ep = null;
	l_deactivate_pending = false;
	l_active = false;
    }
    /* I_Listener.l_reset_Internal_State() */

    
    /**************************************************************************
    **************************************************************************/

    
    // Object state flag (true --> active, false --> inactive)
    private boolean				l_active;

    // NSML startpoint and endpoint object used to form a connection
    private I_Endpoint				l_ep;

    // URL used to connection to this listener
    private String				l_url;
    
    // Event handling information
    private Event_Listener			l_event_listener;
    
    /* synchronization variables to ensure deactivation doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    private I_Ref_Count				l_ref_cnt;
    private boolean				l_deactivate_pending;
}
