/******************************************************************************
I_Startpoint.java

CVS Information:

  $Source: /homes/dsl/Master/cif/CIF-Comm-Java-Current/libraries/cif/comm/I_Startpoint.java,v $
  $Date: 1998/10/13 20:25:47 $
  $Revision: 1.4 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import cif.toolkit.*;
import nexus.*;


/**
 * This class defines the startpoint of a Nexus Simple Message Layer (NSML')
 * connection.  The startpoint is used to send messages to a NSML endpoint.
 */
class I_Startpoint
{
    /**
     * Initialize internal startpoint object state.  To activate a startpoint
     * object, call one of the activate() methods.
     */
    public
    I_Startpoint()
    {
	l_ref_cnt = new I_Ref_Count();
	l_reset_Internal_State();
    }


    /**
     * Enable the startpoint by constructing it from an existing endpoint.
     *
     * @param ep
     *
     *     an active endpoint from which to construct the startpoint
     *
     * @exception cif.toolkit.Object_Active_Exception
     *
     *     Object_Active_Exception is thrown if the object has been
     *     previously activated.
     *
     * @exception cif.toolkit.Bad_Parameter_Exception
     *
     *     Bad_Parameter_Exception is thrown if the supplied endpoint object is
     *     not active.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown endpoint object is not active or is
     *     not capable of having a startpoint bound to it
     */
    public synchronized
    void
    activate(
	I_Endpoint				ep)
    throws
	Object_Active_Exception,
	Bad_Parameter_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Verify that the object is not currently active
	 */
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}

	/*
	 * Extract the Nexus endpoint object out of the endpoint class
	 */
	Endpoint				nx_ep;
	
	try
	{
	    nx_ep = ep.i_get_Nexus_Endpoint();
	}
	catch (Object_Inactive_Exception e)
	{
	    throw new Bad_Parameter_Exception(
		1, "I_Startpoint.activate(): bad endpoint parameter");
	}

	/*
	 * Create a new startpoint from the endpoint
	 */
	nexus.Nexus				nx;

	nx = Comm.acquire_Nexus();
	
	l_nx_sp = nx.startpoint_bind(nx_ep);

	Comm.release_Nexus();

	/*
	 * Update object state to indicate that the startpoint is now active
	 */
	l_active = true;
    }


    /**
     * Enable the startpoint by performing a remote attach request
     *
     * @param hostname
     *
     *     host to submit a remote attach request to
     *
     * @param port
     *
     *     port number to submit the remote attach request to
     *
     * @exception cif.toolkit.Object_Active_Exception
     *
     *     Object_Active_Exception is thrown if the object has been
     *     previously activated.
     *
     * @exception cif.toolkit.Bad_Parameter_Exception
     *
     *     Bad_Parameter_Exception is thrown if port number is bad
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown endpoint object is not active or is
     *     not capable of having a startpoint bound to it
     */
    public synchronized
    void
    activate(
	String					hostname,
	int					port)
    throws
	Object_Active_Exception,
	Bad_Parameter_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Verify that the object is not currently active
	 */
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}

	/*
	 * Create a URL from the supplied information
	 */
	String					url;

	if (port < 0 || port > 65535)
	{
	    throw new Bad_Parameter_Exception(2, "bad port number");
	}
	
	url = "x-nexus://" + hostname + ":" + port;
	
	/*
	 * Create a new startpoint from the endpoint
	 */
	nexus.Nexus				nx;
	nexus.AttachReturn			attach_info;

	nx = Comm.acquire_Nexus();
	attach_info = nx.attach(url);
	Comm.release_Nexus();
	
	if (attach_info.status != 0)
	{
	    throw new Comm_Failure_Exception("attach failed");
	}
	
	l_nx_sp = attach_info.sp;
	

	/*
	 * Update object state to indicate that the startpoint is now active
	 */
	l_active = true;
    }


    /**
     * Extract the startpoint from a Nexus buffer
     *
     * @param nx_buf
     *
     *     Nexus buffer contains a serialized startpoint
     *
     * @exception cif.toolkit.Object_Active_Exception
     *
     *     Object_Active_Exception is thrown if the object has been
     *     previously activated.
     *
     * @exception cif.toolkit.Bad_Parameter_Exception
     *
     *     Bad_Parameter_Exception is thrown if port number is bad
     */
    public synchronized
    void
    activate(
	GetBuffer				nx_buf)
    throws
	Object_Active_Exception,
	Bad_Parameter_Exception
    {
	/*
	 * Verify that the object is not currently active
	 */
	if (l_active)
	{
	    throw new Object_Active_Exception();
	}

	/*
	 * Create a new startpoint from the endpoint
	 */
	try
	{
	    l_nx_sp = nx_buf.get_startpoint();
	}
	catch (Exception e)
	{
	    throw new Bad_Parameter_Exception(
		1, "buffer contains bad startpoint");
	}
	
	/*
	 * Update object state to indicate that the startpoint is now active
	 */
	l_active = true;
    }


    /**
     * Send a message to the endpoint(s) associated with this connection.
     * This method returns as soon as the buffer has been delivered to the
     * underlying protocol module.  In other words, successful execution of
     * this method does not imply that the message has been delivered to the
     * endpoint (or the local operating system).
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the previously sent message may not have been
     *     delivered.
     */
    public
    void
    send(
	PutBuffer				buffer)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active || l_deactivate_pending)
	    {
		throw new Object_Inactive_Exception();
	    }

	    /*
	     * Increment object reference count so that deactivation cannot
	     * occur while we are using the startpoint to perform the send
	     * operation
	     */
	    l_ref_cnt.inc_Count();
	}

	/*
	 * Send 
	 */
	try
	{
	    nexus.Nexus				nx;
	
	    nx = Comm.acquire_Nexus();
	    nx.send_rsr(buffer, l_nx_sp, 0, true);
	    Comm.release_Nexus();
	}
	finally
	{
	    l_ref_cnt.dec_Count();
	}
    }


    public synchronized
    int
    sizeof()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	int					size;
	
	/*
	 * Verify that the object has been activated
	 */
	if (!l_active)
	{
	    throw new Object_Inactive_Exception();
	}

	size = l_nx_sp.sizeof();
	if (size <= 0)
	{
	    throw new Internal_Error(
		"cif.comm.I_Startpoint: startpoint size <= 0");
	}

	return size;
    }
    
    
    public
    void
    put(
	PutBuffer				nx_buf)
    throws
	Object_Inactive_Exception,
	Bad_Parameter_Exception
    {
	/*
	 * Initiate deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    l_deactivate_pending = true;
	}
	// end synchronized(this)
	    
	/*
	 * Wait for blocking operations to complete
	 */
	l_ref_cnt.wait_Until_Zero();

	/*
	 * Finsh deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Since another thread may have already finished the deactivation,
	     * we need to verify that the object is still active
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    /*
	     * Destroy the startpoint
	     */
	    try
	    {
		l_nx_sp.put(nx_buf);
	    }
	    catch (BufferOverrunException e)
	    {
		throw new Bad_Parameter_Exception(
		    1, "cif.comm.I_Startpoint.put(): buffer too small");
	    }

	    /*
	     * Reset the object's internal state
	     */
	    l_reset_Internal_State();
	}
    }
    
    
    /**
     * Disable the startpoint.  Any pending messages are flushed prior to
     * closing the underlying communication channel.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the object has not been
     *     previously activated or has been deactivated.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the previously sent message may not have been
     *     delivered.  This is exception is provided for imformative purposes
     *     only.  The deactivation operation is still performed.
     */
    public synchronized
    void
    deactivate()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception
    {
	/*
	 * Initiate deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Verify that the object has been activated
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    l_deactivate_pending = true;
	}
	// end synchronized(this)
	    
	/*
	 * Wait for blocking operations to complete
	 */
	l_ref_cnt.wait_Until_Zero();

	/*
	 * Finsh deactivation...
	 */
	synchronized(this)
	{
	    /*
	     * Since another thread may have already finished the deactivation,
	     * we need to verify that the object is still active
	     */
	    if (!l_active)
	    {
		throw new Object_Inactive_Exception();
	    }

	    /*
	     * Destroy the startpoint
	     */
	    l_nx_sp.destroy();

	    /*
	     * Reset the object's internal state
	     */
	    l_reset_Internal_State();
	}
	// end synchronized(this)
    }
    

    /**
     * Disable the startpoint.  This method simply calls the deactivate()
     * method in a last ditch effort to properly deactivate the startpoint
     * object before it is deleted.
     */
    public
    void
    finalize()
    {
	try
	{
	    deactivate();
	}
	catch (Exception e)
	{
	}
    }
    /* I_Startpoint.finalize() */

    /**************************************************************************
    **************************************************************************/

    /**
     * I_Startpoint.l_reset_Internal_State()
     */
    private
    void
    l_reset_Internal_State()
    {
	l_nx_sp = null;
	l_deactivate_pending = false;
	l_active = false;
    }
    /* I_Startpoint.l_reset_Internal_State() */

    /**************************************************************************
    **************************************************************************/
    /*
     * Internal data members
     */
    
    // Object state flag (true --> active, false --> inactive)
    private boolean				l_active;

    // Nexus startpoint for sending messages
    private Startpoint				l_nx_sp;

    /* synchronization variables to ensure deactivation doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    private I_Ref_Count				l_ref_cnt;
    private boolean				l_deactivate_pending;
}
