/*
 * TestNexus.java
 *
 * Test of the nexus Java package.
 */

import nexus.*;
import java.net.*;

/*
 * class TestNexus
 */
public class TestNexus
    implements AttachApprovalInterface
{
    /*
     * Private instance variables
     */
    private Nexus		nexus;

    private boolean		run_local_test_arg = false;
    private boolean		attach_server_arg = false;
    private boolean		attach_client_arg = false;
    private String		attach_client_url;
    private boolean		attach_client_leave_server = false;
    private boolean		got_server_shutdown = false;

 
    /*
     * main()
     */
    public static void main (String argv[])
    {
	TestNexus n = new TestNexus();

	n.start(argv);
    } /* main() */

    
    /*
     * start()
     */
    public void start(String args[])
    {
	nexus = new Nexus();

	nexus.activate();

	parse_args(args);

	if (run_local_test_arg)
	{
	    CommTest ct = new CommTest(nexus, this);
	    ct.run();
	}
	
	if (attach_server_arg)
	{
	    attach_server();
	}
	else if (attach_client_arg)
	{
	    attach_client();
	}

	nexus.deactivate();
    } /* start() */


    /*
     * parse_args()
     */
    protected void parse_args(String args[])
    {
	int length = args.length;
	int i;

	i = 0;
	while (i < length)
	{
	    if (args[i].equals("-local"))
	    {
		run_local_test_arg = true;
		i += 1;
	    }
	    else if (args[i].equals("-server"))
	    {
		attach_server_arg = true;
		i += 1;
	    }
	    else if (   (args[i].equals("-client"))
		     && (i+1 < length) )
	    {
		attach_client_arg = true;
		attach_client_url = args[i+1];
		i += 2;
	    }
	    else if (args[i].equals("-leave_server"))
	    {
		attach_client_leave_server = true;
		i += 1;
	    }
	    else
	    {
		i += 1;
	    }
	}
    } /* parse_args() */


    /*
     * attach_server()
     */
    protected void attach_server()
    {
	int port;
	String hostname;
	CommTest commtest;

	commtest = new CommTest(nexus, this);
	
	System.out.println("\nTestNexus.attach_server(): calling allow_attach()");
	port = nexus.allow_attach(0, this, commtest);
	hostname = nexus.attach_hostname();
System.out.println("TestNexus.attach_server(): allow_attach() returned\nTestNexus server listening on x-nexus://"+hostname+":"+port+"/");

	wait_for_server_shutdown();
	
	nexus.disallow_attach(port);
    } /* attach_server() */

    
    /*
     * attach_approval()
     *
     * Method of AttachApprovalInterface
     */
    public AttachReturn attach_approval(Object user_arg,
					String url)
    {
	CommTest commtest;
	EndpointAttr ct_epattr;
	Endpoint ct_ep;
	Startpoint ct_sp;

	commtest = (CommTest) user_arg;

	ct_epattr = nexus.endpointattr_init();
	ct_epattr.set_handler_object(commtest);
        ct_epattr.set_protocol(new TcpInfo());
	ct_ep = nexus.endpoint_init(ct_epattr);
	ct_sp = nexus.startpoint_bind(ct_ep);
	return (new AttachReturn(0, ct_sp));
    } /* attach_approval() */


    /*
     * wait_for_server_shutdown()
     */
    public synchronized void wait_for_server_shutdown()
    {
	while (!got_server_shutdown)
	{
	    try
	    {
		wait();
	    }
	    catch (Exception e) { e.printStackTrace(); }
	}
    } /* wait_for_server_shutdown() */


    /*
     * shutdown_server()
     */
    public synchronized void shutdown_server()
    {
	got_server_shutdown = true;
	try
	{
	    notifyAll();
	}
	catch (Exception e) { e.printStackTrace(); }
    } /* shutdown_server() */

    
    /*
     * attach_client()
     */
    protected void attach_client()
    {
	CommTest commtest;
	AttachReturn attach_return;
	Startpoint server_sp;
	String url;
	
	url = attach_client_url;
	
	attach_return = nexus.attach(url);

	if (attach_return.status != 0)
	{
	    System.out.println(
		"TestNexus.attach_client(): Failed to attach to "+url+".");
		return;
	}
	
	System.out.println(
	    "TestNexus.attach_client(): Attached to \""+url+"\".");

	server_sp = attach_return.sp;

	commtest = new CommTest(nexus, this, server_sp);
	commtest.run();

	if (!attach_client_leave_server)
	{
	    commtest.attach_server_finish(server_sp);
	}

    } /* attach_client() */
    
} /* class TestNexus */


/*
 * class CommTest
 */
class CommTest implements HandlerInterface
{
    private Nexus		nexus;
    private TestNexus		testnexus;
    private EndpointAttr 	epattr;
    private Endpoint 		ep;
    private Startpoint      	local_sp;
    private Startpoint      	server_sp;
    private boolean		reply_done;

    private static final int	RECEIVE_HANDLER = 0;
    private static final int	REPLY_HANDLER = 1;
    private static final int	SERVER_SHUTDOWN_HANDLER = 2;
    
    /*
     * CommTest()
     */
    public CommTest(Nexus nexus, TestNexus testnexus)
    {
	this.nexus = nexus;
	this.testnexus = testnexus;
	this.server_sp = null;
    } 
    public CommTest(Nexus nexus, TestNexus testnexus, Startpoint server_sp)
    {
	this.nexus = nexus;
	this.testnexus = testnexus;
	this.server_sp = server_sp;
    } /* CommTest() */
    

    /*
     * run()
     */
    public void run()
    {
	PutBuffer buffer;

	reply_done = false;

	// create endpoint and bind startpoint to it
	epattr = nexus.endpointattr_init();
	epattr.set_handler_object(this);
	//epattr.set_protocol(new UdpInfo("140.221.9.179", 32598, (byte)0));
	//epattr.set_protocol(new UdpInfo("140.221.14.56", 32598, (byte)0));
	//epattr.set_protocol(new UdpInfo("239.255.255.255", 32598, (byte)0));
	epattr.set_protocol(new TcpInfo());
	ep = nexus.endpoint_init(epattr);
	local_sp = nexus.startpoint_bind(ep);

	// init and pack buffer
	buffer = pack_buffer();

        System.out.println("CommTest.run(): send_rsr to receive handler");
	if (server_sp == null)
	{
	    nexus.send_rsr(buffer, local_sp, RECEIVE_HANDLER, false);
	}
	else
	{
	    nexus.send_rsr(buffer, server_sp, RECEIVE_HANDLER, false);
	}

	// wait for reply
	System.out.println("CommTest.run(): waiting for reply");
	wait_for_reply();
	System.out.println("CommTest.run(): reply received");
    } /* run() */
    

    /*
     * pack_buffer()
     */
    public PutBuffer pack_buffer()
    {
	PutBuffer buffer;
	int size;
	char c = 8;
	short s = 24;
	int i = 42;
	float f = 12.345f;
	double d = 67.89;

	size = (local_sp.sizeof()
		+ nexus.sizeof_char(1)
		+ nexus.sizeof_short(1)
		+ nexus.sizeof_int(1)
		+ nexus.sizeof_float(1)
		+ nexus.sizeof_double(1));
	buffer = nexus.buffer_init(size, 0);

	try
	{
	    local_sp.put(buffer);
	    buffer.put_char(c);
	    buffer.put_short(s);
	    buffer.put_int(i);
	    buffer.put_float(f);
	    buffer.put_double(d);

            System.out.println("CommTest.pack_buffer(): put startpoint");
            System.out.println("CommTest.pack_buffer(): put c="+(int)c+", s="+s+", i="+i+", f="+f+", d="+d);
	}
	catch (Exception e) { e.printStackTrace(); }

	return buffer;
    } /* pack_buffer() */


    /*
     * wait_for_reply()
     */
    private synchronized void wait_for_reply()
    {
	while (!reply_done)
	{
	    try
	    {
		wait();
	    }
	    catch (Exception e) { e.printStackTrace(); }
	}
    } /* wait_for_reply() */


    /*
     * notify_of_reply()
     */
    private synchronized void notify_of_reply()
    {
	reply_done = true;
	try
	{
	    notifyAll();
	}
	catch (Exception e) { e.printStackTrace(); }
    } /* notify_of_reply() */
    

    /*
     * receive_handler()
     */
    public void receive_handler(GetBuffer buffer)   
    {
	char c;
	short s;
	int i;
	float f;
	double d;
	int destroy_context;
	Startpoint reply_sp;
	PutBuffer reply_buffer;

	System.out.println("CommTest.receive_handler(): thread="+Thread.currentThread().getName());
	
	try
	{
	    reply_sp = new Startpoint(nexus, buffer);
	    System.out.println("CommTest.receive_handler(): got reply_sp");

	    c = buffer.get_char();
	    s = buffer.get_short();
	    i = buffer.get_int();
	    f = buffer.get_float();
	    d = buffer.get_double();
	    System.out.println("CommTest.receive_handler(): got c="+(int)c+", s="+s+", i="+i+", f="+f+", d="+d);

	    reply_buffer = nexus.buffer_init(0, 0);
	    System.out.println("CommTest.receive_handler(): send_rsr to reply handler");
	    nexus.send_rsr(reply_buffer, reply_sp, REPLY_HANDLER, false);
	}
	catch (Exception e) { e.printStackTrace(); }
    } /* receive_handler() */
    

    /*
     * reply_handler()
     */
    public void reply_handler()
    {
	System.out.println("CommTest.remote_reply_handler(): thread="+Thread.currentThread().getName());
	notify_of_reply(); 
    } /* reply_handler() */
    
	
    /*
     * server_shutdown_handler()
     */
    public void server_shutdown_handler()
    {
	System.out.println("CommTest.server_shutdown_handler(): thread="+Thread.currentThread().getName());
	testnexus.shutdown_server();
    } /* server_shutdown_handler() */
    

    /*
     * attach_server_finish()
     */
    public void attach_server_finish(Startpoint server_sp)
    {
	PutBuffer buffer;

	try
	{
	    buffer = nexus.buffer_init(0, 0);	
	    System.out.println("CommTest.attach_server_finish(): send_rsr to server shutdown handler");
	    nexus.send_rsr(buffer, server_sp, SERVER_SHUTDOWN_HANDLER, false);
	}
	catch (Exception e) { e.printStackTrace(); }

    } /* attach_server_finish() */
    

    /*
     * invoke_handler()
     *
     * This is part of the HandlerInterface.
     * It is the entry point for any handler invocations into this object.
     */
    public void invoke_handler(Endpoint ep,
			       GetBuffer buffer,
			       int handler_id)
    {
	switch (handler_id)
	{
	case 0:
	    receive_handler(buffer);
	    break;
	case 1:
	    reply_handler();
	    break;
	case 2:
	    server_shutdown_handler();
	    break;
	default:
	    System.out.println("CommTest.invoke_handler(): Error: handler not recognized within object");
	    break;
	}
    } /* invoke_handler() */

} /* class CommTest */

