#include <stdio.h>
#include <string.h>
#include <globus_nexus.h>

/* Nexus Handler Functions */
static void receive_handler(nexus_endpoint_t *endpoint, 
                            nexus_buffer_t *buffer, 
                            nexus_bool_t is_non_threaded_handler);
static void dummy_handler(nexus_endpoint_t *endpoint, 
                          nexus_buffer_t *buffer,
                          nexus_bool_t is_non_threaded_handler);
static void server_shutdown_handler(nexus_endpoint_t *endpoint, 
                                    nexus_buffer_t *buffer,
                                    nexus_bool_t is_non_threaded_handler);

/* Nexus Handler Functions Table */
#define RECEIVE_HANDLER		0
#define REPLY_HANDLER		1	
#define SERVER_SHUTDOWN_HANDLER	2

static nexus_handler_t handlers[] = 
{
    {NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) receive_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) dummy_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) server_shutdown_handler}
};


/* Data Structures */
typedef struct
{
    /* for mutually exclusive access to monitor */
    nexus_mutex_t mutex;

    /* a monitor condition variable and flag */
    nexus_cond_t  cond;
    nexus_bool_t  done;
    nexus_bool_t  connection_established;
} monitor_t;


/* Global Variables */
static nexus_endpointattr_t	GlobalEndpointAttr;
static nexus_endpoint_t 	GlobalEndpoint;
static monitor_t		Monitor;


/* Local Functions */
static int attach_requested(void *arg, char *url, nexus_startpoint_t *sp);
static void allow_attachments(unsigned short *attach_port, char *attach_url);


/*
 * main()
 */ 
int main(int argc, char **argv)
{
  int i, ch;
  nexus_buffer_t buffer;
  char attach_url[1024];
  unsigned short attach_port;

  globus_module_activate(GLOBUS_NEXUS_MODULE);

  /* Initialization */
  nexus_endpointattr_init(&GlobalEndpointAttr);

#if 0
  nexus_endpointattr_set_protocol(&GlobalEndpointAttr,
                                  NEXUS_PROTO_TYPE_UDP,
                                  NULL, 0);
#endif

  nexus_endpointattr_set_handler_table(&GlobalEndpointAttr, handlers,
    sizeof(handlers)/sizeof(nexus_handler_t));
  nexus_endpoint_init(&GlobalEndpoint, &GlobalEndpointAttr);
  nexus_endpoint_set_user_pointer(&GlobalEndpoint, (void *) (&Monitor));

  /* init monitor */
  nexus_mutex_init(&(Monitor.mutex), (nexus_mutexattr_t *) NULL);
  nexus_cond_init(&(Monitor.cond), (nexus_condattr_t *) NULL);
  nexus_mutex_lock(&(Monitor.mutex));
  Monitor.done = NEXUS_FALSE;
  Monitor.connection_established = NEXUS_FALSE;
  nexus_mutex_unlock(&(Monitor.mutex));

  /* allowing attachment */
  allow_attachments(&attach_port, attach_url);

  /* wait on server_shutdown_handler invocation */
  nexus_mutex_lock(&(Monitor.mutex));
  while (!(Monitor.done))
  {
    nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));
  } /* endwhile */
  nexus_mutex_unlock(&(Monitor.mutex));

  /* Shutdown */
  nexus_printf("test_server: calling nexus_disallow_attach()\n");
  nexus_disallow_attach(attach_port);
  nexus_printf("test_server: nexus_disallow_attach returned()\n");

  /* Clean-Up */
  nexus_mutex_destroy(&(Monitor.mutex));
  nexus_cond_destroy(&(Monitor.cond));
  nexus_endpoint_destroy(&GlobalEndpoint);
  nexus_endpointattr_destroy(&GlobalEndpointAttr);

  globus_module_deactivate(GLOBUS_NEXUS_MODULE);

  exit(0);

} /* end main() */

/*
 * attach_requested()
 */
static int attach_requested(void *arg,
                            char *url,
                            nexus_startpoint_t *sp)
{
  int rc;

  nexus_printf("attach_requested(): entering\n");

  nexus_mutex_lock(&(Monitor.mutex));
    if (Monitor.connection_established)
    {
      /* another single_rsr_client has already connected to us */
      rc = 1;
    }
    else
    {
      /* this is the first single_rsr_client */
      rc = 0;
      Monitor.connection_established = NEXUS_TRUE;
    } /* endif */
  nexus_mutex_unlock(&(Monitor.mutex));

  if (rc == 0)
    nexus_startpoint_bind(sp, &GlobalEndpoint);

  nexus_printf("attach_requested(): exiting\n");
  return rc;
} /* end attach_requested() */

/*
 * allow_attachments()
 */
static void allow_attachments(unsigned short *attach_port,
                              char *attach_url)
{
  int rc;
  char *attach_host;

  /* allow other Nexus programs to attach to us */
  *attach_port = 0;
  nexus_printf("test_server: calling nexus_allow_attach()\n");
  rc = nexus_allow_attach(attach_port, &attach_host,
                            &attach_requested, NULL);
  if (rc != 0)
  {
     globus_fatal("test_server: nexus_allow_attach() failed: rc=%d\n", rc);
  } /* endif */

  sprintf(attach_url, "x-nexus://%s:%hu/", attach_host, *attach_port);
  nexus_printf("test_server: listening on %s\n", attach_url);
} /* end allow_attachments() */

/*
 * receive_handler()
 */
static void receive_handler(nexus_endpoint_t *endpoint, 
			    nexus_buffer_t *buffer, 
			    nexus_bool_t is_non_threaded_handler)
{
  nexus_startpoint_t client_sp;

  char c;
  short s;
  int i;
  float f;
  double d;

  nexus_printf("test_server: entering receive handler\n");

  /* extracting client startpoint from buffer */
  nexus_get_startpoint(buffer, &client_sp, 1);
  nexus_printf("test_server: got client sp\n");

  /* extracting test data from buffer */
  nexus_get_char(buffer, &c, 1);
  nexus_get_short(buffer, &s, 1);
  nexus_get_int(buffer, &i, 1);
  nexus_get_float(buffer, &f, 1);
  nexus_get_double(buffer, &d, 1);

  nexus_printf("test_server: got c=%d, s=%d, i=%d, f=%f, d=%f\n",
			c, s, i, f, d);

  /* reply to client */
  nexus_buffer_reset(buffer, 0, 0);
  nexus_printf("test_server: calling send_rsr - REPLY HANDLER\n");
  nexus_send_rsr(buffer, 
                 &client_sp, 
                 REPLY_HANDLER, 
                 NEXUS_TRUE, 
                 NEXUS_FALSE);

  /* destroy client sp */
  nexus_startpoint_destroy(&client_sp);

  nexus_printf("test_server: exiting receive handler\n");

} /* receive_handler() */

/*
 * dummy_handler()
 */
static void dummy_handler(nexus_endpoint_t *endpoint, 
                          nexus_buffer_t *buffer,
                          nexus_bool_t is_non_threaded_handler)
{
  nexus_printf("talk_server: dummy handler - should not be here");

  /* clean-up */
  nexus_buffer_destroy(buffer);
} /* dummy_handler() */

/*
 * server_shutdown_handler()
 */
static void server_shutdown_handler(nexus_endpoint_t *endpoint, 
				    nexus_buffer_t *buffer,
				    nexus_bool_t is_non_threaded_handler)
{
  nexus_printf("test_server: entering server shutdown handler\n");

  /* clean-up */
  nexus_buffer_destroy(buffer);

  /* allow server to run to termination */
  nexus_mutex_lock(&(Monitor.mutex));
  Monitor.done = NEXUS_TRUE;
  nexus_cond_signal(&(Monitor.cond));
  nexus_mutex_unlock(&(Monitor.mutex));      

  nexus_printf("test_server: exiting server shutdown handler\n");

} /* server_shutdown_handler() */

